/*
    This sample illustrates how to a save point cloud to a plain text file.

    One single image is acquired. The 3D and intensity data are written to a text file.

    The Point Cloud Library data format (.pcd) is used: http://pointclouds.org/documentation/tutorials/pcd_file_format.php.

    You can open .pcd files using the free CloudCompare software: http://www.danielgm.net/cc/

    A more efficient way to save point clouds to file is to use the Point Cloud Library (PCL).
    The PCL's io module allows to save point clouds in binary .pcd or .ply files.

*/

#include <string>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ConsumerImplHelper/BlazeCamera.h>
#include <GenTL/PFNC.h>

using namespace GenTLConsumerImplHelper;
using namespace GenApi;

void WritePcdHeader(std::ostream& o, size_t width, size_t height)
{
    o << "# .PCD v0.7 - Point Cloud Data file format" << '\n';
    o << "VERSION 0.7" << '\n';
    o << "FIELDS x y z rgb" << '\n';
    o << "SIZE 4 4 4 4\n";
    o << "TYPE F F F F\n";
    o << "COUNT 1 1 1 1\n";
    o << "WIDTH " << width << '\n';
    o << "HEIGHT " << height << '\n';
    o << "VIEWPOINT 0 0 0 1 0 0 0\n";
    o << "POINTS " << width * height << '\n';
    o << "DATA ASCII" << std::endl;
}

bool SavePointCloud(const BufferParts& parts, const char* fileName)
{
    if (parts.empty())
    {
        std::cerr << "No valid image data." << std::endl;
        return false;
    }

    // If the point cloud is enabled, the first part always contains the point cloud data.
    if (parts[0].dataFormat != PFNC_Coord3D_ABC32f)
    {
        std::cerr << "Unexpected data format for the first image part. Coord3D_ABC32f is expected." << std::endl;
        return false;
    }

    if (parts.size() < 2)
    {
        std::cerr << "Unexpected configuration. Point cloud and intensity image must be enabled" << std::endl;
        return false;
    }
    if (parts[1].dataFormat != PFNC_Mono16)
    {
        std::cerr << "Unexpected data format for the second image part. Mono 16 is expected." << std::endl;
        return false;
    }

    std::ofstream o(fileName);
    if (!o)
    {
        std::cerr << "Error:\tFailed to create file " << fileName << std::endl;
        return false;
    }

    std::cout << "Writing point cloud to file " << fileName << "...";
    CBlazeCamera::Coord3D* pPoint = (CBlazeCamera::Coord3D*) parts[0].pData;
    uint16_t* pIntensity = (uint16_t*)parts[1].pData;
    const size_t nPixel = parts[0].width * parts[0].height;

    WritePcdHeader(o, parts[0].width, parts[0].height);

    for (size_t i = 0; i < nPixel; ++i)
    {
        // Check if there are valid 3D coordinates for that pixel.
        if (pPoint->z != 0)
        {
            o.precision(0);  // Coordinates will be written as whole numbers.

            // Write the coordinates of the next point. Note: Since the coordinate system
            // used by the CloudCompare tool is different from the one used by the blaze camera, 
            // we apply a 180-degree rotation around the x-axis by writing the negative 
            // values of the y and z coordinates.
            o << std::fixed << pPoint->x << ' ' << -pPoint->y << ' ' << -pPoint->z;

            // Save the intensity as an RGB value.
            uint8_t gray = *pIntensity >> 8;
            uint32_t rgb = (uint32_t)gray << 16 | (uint32_t)gray << 8 | (uint32_t)gray;
            // The point cloud library data format represents RGB values as floats. 
            float fRgb = *(float*)&rgb;
            o.unsetf(std::ios_base::floatfield); // Switch to default float formatting
            o.precision(9); // Intensity information will be written with highest precision.
            o << ' ' << fRgb << '\n';
        }
        else
        {
            o << "nan nan nan 0\n";
        }
        pPoint++;
        pIntensity++;
    }
    o.close();
    std::cout << "done." << std::endl;
    return true;
}


int main(int argc, char* argv[])
{
    int exitCode = EXIT_FAILURE;
    const char* fileName = argc > 1 ? argv[1] : "points.pcd";

    CBlazeCamera camera;
    try
    {
        CBlazeCamera::InitProducer();
        camera.OpenFirstCamera();

        CEnumerationPtr ptrComponentSelector = camera.GetParameter("ComponentSelector");
        CBooleanPtr ptrComponentEnable = camera.GetParameter("ComponentEnable");
        CEnumerationPtr ptrPixelFormat = camera.GetParameter("PixelFormat");

        // Parameterize the camera to send 3D coordinates and intensity data
        ptrComponentSelector->FromString("Range");
        ptrComponentEnable->SetValue(true);
        ptrPixelFormat->FromString("Coord3D_ABC32f");

        ptrComponentSelector->FromString("Intensity");
        ptrComponentEnable->SetValue(true);
        ptrPixelFormat->FromString("Mono16");

        // Configure the value used for identifiyng missing depth information.
        GenApi::CEnumerationPtr(camera.GetParameter("Scan3dCoordinateSelector"))->FromString("CoordinateC"); // Choose z-axis
        // z-coordinate will be set to this value if there is no valid depth data available
        GenApi::CFloatPtr(camera.GetParameter("Scan3dInvalidDataValue"))->SetValue(0.0);

        // Acquire one single image
        BufferParts parts;
        GrabResultPtr ptrGrabResult = camera.GrabSingleImage(1000, &parts);

        if (ptrGrabResult->status == GrabResult::Ok)
        {
            SavePointCloud(parts, fileName);
        }
        else
        {
            std::cerr << "Failed to grab an image." << std::endl;
        }

        camera.Close();
        exitCode = EXIT_SUCCESS;
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used 
        // to check if the device is still connected.
        if (camera.IsOpen() && !camera.IsConnected())
        {
            std::cerr << "Camera has been removed." << std::endl;
        }
    }

    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}

